package App::ListBase;

use strict;
use warnings;

use Carp;
use Data::Dumper;
use Template;
use Storable qw(dclone);

use vars qw(%c $session $cgi @ISA);
*c = \*app::c;
*session = \$app::session;
*cgi = \$app::cgi;

sub new
{
  my $class = shift;
  
  my $this = {
    web         => undef,
    baseurl     => undef,

    listdef_name	=> undef,
    datasource		=> undef,
    name				=> undef,
    listdef			=> undef,
    errmsg			=> [],
    params			=> {},
    stash				=> {},
    action			=> '',
    buttonstate    => '',
    
    @_,
    
    id					=> undef,
    v					=> {},
    rowstart			=> 0,
    pagesize			=> 10,
    rowend			=> 1,
    rowcount			=> undef,
    ordercol			=> undef,
    orderdir			=> undef,
    firstrun			=> 0,
  };
  bless($this, $class);
  
  $this->init;
  
  $this;
}

sub name { $_[0]->{listdef_name}; }
sub datasource { $_[0]->{datasource}; }
sub listdef { $_[0]->{listdef}; }

sub init
{
  my $this = shift;
  return if $this->{init_done};
  $this->reinit(@_);
  $this->{init_done} = 1;
}

sub reinit
{
	my $this = shift;
	
	my $listdef;
	if ( $this->{listdef_name} )
	{
		unless ( $c{listdef}->{$this->{listdef_name}} )
		{
			confess "LISTDEFROOT not defined" unless $c{LISTDEFROOT};
			for (split /:/,$c{LISTDEFROOT})
			{
				my $path = "${_}/$this->{listdef_name}";
				if ( -r $path )
				{
					my $xp = $this->{web}->read_structure_from_file($path);
					confess "incorrect listdef file: $path" unless ref($xp) eq "HASH" && keys %$xp;
					$c{listdef}->{$this->{listdef_name}} = $xp;
					last;
				}
			}
		}
		$listdef = $c{listdef}->{$this->{listdef_name}}
			|| confess "missing listdef file: $this->{listdef_name} in dirs $c{LISTDEFROOT}";
		$this->{listdef} = dclone($listdef); # take copy so we can update {cols} and so on
	}

	($this->{name} ||= $this->{listdef}->{title}||'List') =~ s/[^A-Za-z0-9]//g;
	my $id = $this->{id}	= 'li'.(($c{'list_unique'}||=1)++);
	my $view = $this->{listdef}->{view};

	# load any helper base classes
	if ( my $hc = $this->{listdef}->{'helperclasses'} )
	{
		for my $module (@$hc)
		{
			# check if loaded first?
			eval "require $module"
				or confess "list $this->{listdef_name}: error loading helper class $module: $@";
			unshift @ISA, $module;
		}
	}

	# form field post values
	my $fv = ( $this->{iv} ||= {%{$cgi->Vars}} );
	$this->{firstrun} = defined $fv->{"${id}_rowstart"} ? 0 : 1;

	# parse list control field values
	$this->{rowstart}	= $fv->{"${id}_rowstart"} || $view->{'rowstart'} || 0;
	$this->{pagesize} = $fv->{"${id}_page_size"} || $view->{'pagesize'} || 10;
	$this->{ordercol} = defb( $fv->{"${id}_ordercol"}, $this->{listdef}->{orderdefault}->{col} );
	$this->{orderdir} = lc defb( $fv->{"${id}_orderdir"}, $this->{listdef}->{orderdefault}->{dir} );

	# call any helper init methods
	for my $fn ( @{$this->{listdef}->{initmethods}} )
	{
		confess "list $this->{listdef_name}: unknown init method: $fn" unless $this->can($fn);
		$this->$fn;
	}

	$this->post_handler unless $this->{init_done};
	
	1;
}

sub sis { $_[0] = $_[1] if defined $_[1]; }
sub sus { $_[0] = $_[1] unless defined $_[1]; }
sub defb { return def(@_) || ''; }
sub def { for (@_) { return $_ if defined($_); } undef; }

sub trim
{
	my $s = shift||'';
	$s =~ s/^\s+//;
	$s =~ s/\s+$//;
	return $s;
}

=head2 fv - get/set form field value

1 - type : sf (search field prefix), if (info field prefix), blank (no prefix)
2 - field name
3 - value (optional) sets field value

=cut

sub fv
{
	my $this = shift;
	my $type = shift || 'sf';
	my $field = shift || return '';
	my $value = shift;
	$field = $this->{id}.'_'.$type.'_'.$field if $type;
	$this->{iv}->{$field} = $value if defined $value;
	$this->{iv}->{$field};
}

sub add_msg
{
	my $this = shift;
	push @{$this->{errmsg}}, @_;
}

sub set_msg
{
	my $this = shift;
	$this->{errmsg} = [ @_ ];
}

sub post_handler
{
	my $this = shift;

	# handle special action	
	my $action = $this->{iv}->{'action_'.$this->{id}}||'';
	if ( $action =~ m/^download/ ) # export to Excel
	{
		$this->{action} = 'export_excel';
	}
	
	# handle button click
	my $button = $this->{iv}->{button}||$this->{iv}->{button_name}||'';
	my $button_value = $this->{iv}->{button_value}||'';
	return 1 unless $button;

	my $id = $this->{'id'};
	my $listdef = $this->{'listdef'};

	if ( $button eq "${id}_nav" )
	{
		$this->{rowstart} = $button_value;
	}
	elsif ( $button eq "${id}_sort" )
	{
		my $col = $button_value;
		if  ( $this->{ordercol} ne $col ) # new sort col
		{
			$this->{ordercol} = $col;
			$this->{orderdir} = 'asc';
		}
		else # same sort col
		{
			my %dirmap = ( $col eq ($listdef->{orderdefault}->{col}||'') && lc ($listdef->{orderdefault}->{dir}||'') eq 'desc')
				? ( '' => 'desc', 'desc' => 'asc', 'asc' => '' ) # special case for default desc -> asc
				: ( '' => 'asc', 'asc' => 'desc', 'desc' => '' );
			$this->{orderdir} = $dirmap{$this->{orderdir}}||'';
			if ( $this->{'orderdir'} eq '' ) # turn off, back to default sort
			{
				$this->{ordercol} = $this->{listdef}->{orderdefault}->{col}||'';
				$this->{orderdir} = lc $this->{listdef}->{orderdefault}->{dir}||'';
			}
			else
			{
				$this->{ordercol} = $col;
			}
		}
	}

	1;
}

sub init_datasource
{
	my $this = shift;
	
	my $ds = $this->{datasource} = {};

	# sql_select
	
	1;
}

sub init_display
{
	my $this = shift;

	1;
}

sub fetch_datasource_page
{
	my $this = shift;

	my $listdef = $this->{listdef} || confess;
	my $listds = $listdef->{datasource} || confess;
	my $ds = ($this->{datasource} ||= {});
	my $fn;

	$ds->{rowstart}	= $this->{rowstart}||1;
	$ds->{rowstart}--;
	$ds->{pagesize}	= $this->{pagesize}||10;
	$ds->{dispcols}	= [];
	$ds->{sqlcols}		= [];
	$ds->{allcols}		= [];

	for ( @{$listdef->{cols}} )
	{
		push @{$ds->{allcols}}, $_;
		push @{$ds->{dispcols}}, $_;
		push @{$ds->{sqlcols}}, $_ unless $listdef->{colscalc}->{$_};
	}
	for ( @{$listdef->{cols2}} )
	{
		push @{$ds->{allcols}}, $_;
		push @{$ds->{sqlcols}}, $_ unless $listdef->{colscalc}->{$_};
	}

	my ($db, $dbrs);

	my $order_by;
	if ( my $os = $listdef->{orderspecial}->{$this->{ordercol}} )
	{
		if ($#$os > -1) # empty array means not sortable
		{
			my @ordertmp;
			my $first = 1;
			my $defdir = lc($os->[0]->[1] || 'asc');
			my %dirmap = ( $defdir eq 'desc')
				? ( '' => 'desc', 'desc' => 'asc', 'asc' => '' ) # special case for default desc -> asc
				: ( '' => 'asc', 'asc' => 'desc', 'desc' => '' );
			for my $osptr (@$os)
			{
				my $osfield = $osptr->[0] || confess;
				my $osdir = lc($osptr->[1] || 'asc');
				if ( $first && $this->{orderdir} eq 'desc' )
				{
					$osdir = ($osdir eq 'desc') ? 'asc' : 'desc';
				}
				push @ordertmp, $osfield . ' ' . $osdir;
			}
			#push @order, 'ORDER BY ' . join(', ',@ordertmp);
			$order_by = join(', ',@ordertmp);
		}
	}
	elsif ( $listdef->{ordernone}->{$this->{ordercol}} )
	{
		; # no sort
	}
	else
	{
		#push @order, 'ORDER BY '.$this->{ordercol}.' '.$this->{orderdir} if $this->{ordercol};
		$order_by = $this->{ordercol}.' '.$this->{orderdir} if $this->{ordercol};
	}

	# set up datasource elements		
	$ds->{table} = $listds->{table};
	$ds->{numcols} = $#{$ds->{dispcols}};
	$ds->{keycol} = -1;
	$ds->{numrows} = 0;
	$ds->{name2colno} = {};
	my $i;

	# get data page
	if ( $this->{firstrun} && ($listdef->{view}->{autorun}||0) == 0 ) # empty list on first run
	{
	}
	elsif ( ($listds->{type}||'') eq 'table' ) # retrieve page
	{
    $db = $c{schema}->resultset(ucfirst($ds->{table}));
		$dbrs = $db->result_source;

		# build search clause
		my $from = $listds->{table};
		my $where = '1=1';
		my @where;
		# search fields
		for my $field ( @{$listdef->{search}||[]}, @{$listdef->{search2}||[]} )
		{
			my $val = $this->{iv}->{$this->{id} .'_sf_' .$field} || '';
			next unless $val;

			# calculated search using helper functions, may add joins and where clauses
			if ( $fn = $listdef->{searchcalc}->{$field} )
			{
				my @arg = ( iv => $this->{iv}, field => $field, value => $val, wherestrptr => \$where, whereptr => \@where, dbptr => $db, fromptr => \$from );
				if ( $this->can($fn) ) # method name
				{
					$this->$fn(@arg);
				}
				elsif ( defined &{$fn} ) # function pointer
				{
					no strict 'refs';
					&{$fn}($this, @arg);
					use strict 'refs';
				}
				else
				{
					confess "list $this->{listdef_name}: unknown searchcalc function $fn for field $field";
				}
				next;
			}

			# normal search on database field
			my $type = uc($dbrs->column_info($field)->{data_type});
			if ($type == 'CHAR' || $type == 'VARCHAR' ) # 1 CHAR, 12 VARCHAR
			{
				push @where, { $field => { 'like', $val.'%' } }; # match start of field
			}
			elsif ($type == 'DATE' ) # 9 DATE
			{
				push @where, { $field => $val };
			}
			elsif ($type == 'DATETIME' ) # 10 DATETIME
			{
				push @where, { $field => $val };
			}
			elsif ($type == 'ENUM' )
			{
			  push @where, { $field => $val };
		  }
			else # INT, DOUBLE, DATETIME, TIMESTAMP
			{
				push @where, { $field => $val };
			}
		}

		# any extra searches or joins needed
		for $fn ( @{$listdef->{searchextra}||[]} )
		{
			my @arg = ( iv => $this->{iv}, wherestrptr => \$where, whereptr => \@where, dbptr => $db, fromptr => \$from );
			if ( $this->can($fn) ) # method name
			{
				$this->$fn(@arg);
			}
			elsif ( defined &{$fn} ) # function pointer
			{
				no strict 'refs';
				&{$fn}($this, @arg);
				use strict 'refs';
			}
			else
			{
				confess "list $this->{listdef_name}: unknown searchextra function $fn";
			}
		}

		# get all pages of data if exporting to excel
		if ( $this->{action} =~ m/^export_/ )
		{
			$ds->{rowstart} = 0;
			$ds->{pagesize} = 0; # $c{dbuser}->download_limit;
		}

		# get page of data
		my $rs = $db->search(
		  \@where,
		  {
		    page => 1 + ($ds->{rowstart} / $ds->{pagesize}),
		    rows => $ds->{pagesize},
		    #select => [ 'id', 'name' ],
		    select => $ds->{sqlcols},
		    order_by => $order_by,
		    #group_by => [],
		  }
	  );
		unless ($rs)
		{
			$this->add_msg("List: Sql error: ");#TODO .$db->sql_error);
			return 0;
		}
#$DB::single = 1;
		# copy page to datasource
		$i = 0;
		for my $rp ( $rs->all )
		{
			# optional method after fetch row
			if ( my $fn = $listdef->{on_after_fetchrow} )
			{
				if ( $this->can($fn) )
				{
					my @arg = ( iv => $this->{iv}, dbptr => $db, rp => $rp );
					$this->$fn(@arg);
				}
			}

			my @a;
			for (@{$ds->{allcols}})
			{
				my $calcmeth = $listdef->{colscalc}->{$_};
				if ( $calcmeth )
				{
					confess "unknown calc method: $calcmeth" unless $this->can($calcmeth);
					push @a, $this->$calcmeth(field => $_, rp => $rp);
				}
				else
				{
					push @a, $rp->get_column($_);
				}
			}
			#
			$ds->{row}->[$i] = \@a;
			$ds->{rowextra}->[$i] = ''; #'popup details for '.$rp->get_column($ds->{allcols}->[0]);
			$i ++;
		}
		# get total number of found rows
	  #$ds->{numrows} = $rs->search({}, { select => [ { count => { distinct => 'id' } } ], as => ['count'] })->next->get_column('count');
    $ds->{numrows} = $db->search( \@where, { select => [ { count => { distinct => 'id' } } ], as => ['count'] })->first->get_column('count');

		$ds->{numrows} = $i unless defined $ds->{numrows};
	}
	else
	{
		confess "unknown listdef type: $listds->{type}";
	}

	# map field names to colnos
	for $i ( 0 .. $#{$ds->{allcols}} )
	{
		$ds->{keycol} = $i if $ds->{allcols}->[$i] eq 'id'; # TODO need alternative if primary key field is not 'id'
		$ds->{name2colno}->{$ds->{allcols}->[$i]} = $i;
	}

	# limit to bounds from rowcount
	
	$ds;
}

sub display
{
	my $this = shift;
	my $param = shift;
	confess "expected hashref" if $param && ref $param ne 'HASH';
	$param ||= {};

	$this->init_datasource;

	$this->init_display;

	my $ds = $this->fetch_datasource_page;

  my $template = Template->new({
		INCLUDE_PATH => "$c{TEMPLATE_INCLUDE_PATH}",
		PLUGINS => { W => "App::Widget" },
   });

=head1

search fields, with values
info options
page no
page size
rowcount
sql context to session
sort col U/D

#http://preview.sugarondemand.com/will355/index.php?action=index&query=true&module=Contacts&button=Search&Contacts_CONTACT_offset=40

#http://preview.sugarondemand.com/will355/index.php?action=DetailView&module=Contacts&record=63ce9523-104d-9591-021c-43a31217c351&offset=43&stamp=1139653754034945500

state : list->view or view
=cut

	my $id = $this->{id};
	my $download_limit = 0; # $c{'dbuser'}->download_limit || 0 ;

	# work out row page
	my $rowcount = int($ds->{numrows}||0);
	$rowcount = 0 if $rowcount < 0;
	my $rowstart = int($this->{rowstart}||1);
	$rowstart = $rowcount if $rowstart > $rowcount;
	my $pagesize = int($this->{pagesize}||10);
	$rowstart = $rowstart - (($rowstart -1) % $pagesize); # adjust to page boundary
	$rowstart = 0 if $rowstart < 0;
	my $rowend = $rowstart + $pagesize -1;
	$rowend = $rowcount if $rowend > $rowcount;
	my $pageprev = $rowstart - $pagesize;
	$pageprev = ($rowcount ? 1 : 0) if $pageprev < 1;
	my $pagelast = $rowcount ? ($rowcount - (($rowcount - 1) % $pagesize)) : 0;
	my $pagenext = $rowstart + $pagesize;
	$pagenext = $pagelast if $pagenext > $pagelast;

  $c{v}->{session_id} = $session->id;

#$DB::single = 1;
	# merge to template
	my $p = {
	  # TODO should come from ::Web
	  baseurl     => ($this->{baseurl}||confess 'expected baseurl'),
	  session_id  => $session->id,
	  sess_user   => $session->param('user'),

		web         => $this->{web},

		list				=> $this,
		listid			=> $this->{id},
		listname		=> $this->{name},
		#c					  => \%c,
		#p					=> \%main::PARAM,
		fv					=> $this->{iv},
		ds					=> $this->{datasource},
		buttons			=> $this->{buttons},
		buttons2			=> $this->{buttons2},
		buttonstate    => $this->{buttonstate},

		ld					=> $this->{listdef},
		table       => ($this->{listdef}->{datasource}->{table}||''),
		#linkexport		=> q{javascript:alert('export');},
		linkexport		=> 'export', # #qq{javascript:action_screen('action_${id}','download;msexcel','Download');},
		linkrowstart 	=> qq{javascript:btnclick('${id}_nav',1);},
		linkrowprev		=> qq{javascript:btnclick('${id}_nav',$pageprev);},
		linkrownext		=> qq{javascript:btnclick('${id}_nav',$pagenext);},
		linkrowend		=> qq{javascript:btnclick('${id}_nav',$pagelast);},

		linkview			=> $this->{viewurl},
		download_limit => $download_limit,

		rowstart			=> $rowstart,
		rowcount			=> $rowcount,
		pagenumbers		=> "($rowstart - $rowend of $rowcount)",
		pagesize			=> $pagesize,
		ordercol			=> $this->{ordercol},
		orderdir			=> $this->{orderdir},
		
		list_error_msg	=> (join('<br>',@{$this->{errmsg}})),

		%$param,
		};

	if ( $this->{action} eq 'export_excel' )
	{
		$template->process("list/export.tt2", $p);
		exit(0);
	}

	if ( defined wantarray )
	{
		my $output;
	   $template->process("list/list.tt2", $p, \$output)
			|| confess "Template error: ".$template->error()."\n" ;
		return \$output;
	}
	else
	{
	   $template->process("list/list.tt2", $p)
			|| confess "Template error: ".$template->error()."\n" ;
		return 1;
	}
}

sub display_show
{
	my $this = shift;
	$this->display(@_);
	1;
}

sub display_ptr
{
	my $this = shift;
	my $strptr = $this->display(@_);
	$strptr;
}

sub display_str
{
	my $this = shift;
	my $strptr = $this->display(@_);
	$$strptr;
}

=head2 method, methinfo - safely call a method from a template

=cut

sub method
{
	my $this = shift;
	my $method = shift || confess 'missing method value';
	$this->can($method) or confess 'missing method: '.$method;
	return $this->$method(@_);
}

=head2 row_value_join - join field values from specified row no to give a string

either

rp - pointer to hash of field name => value

or

rowno - number of row in datasource

datasource - optional datasource, defaults to current

=cut

sub rv
{
	my $this = shift;
	my %arg = @_;
	my $rowno = $arg{rowno} || return '';
	$rowno--; # list rows from 1..10, data stored 0..9
	my $field = $arg{field} || confess;
	my $ds = $arg{datasource} || $this->{datasource} || confess;

	my $row = $ds->{row}->[$rowno] || return '';
	my $colno = $ds->{name2colno}->{$field};
	confess "rv: unknown column $field - check cols2 in listdef $this->{listdef_name}" unless defined $colno;
	return $row->[$colno];
}

sub row_value_join
{
	my $this = shift;
	my %arg = @_;
	my $fields = $arg{fields} || [];
	my $sep = $arg{separator} || ', ';
	my $rp = $arg{rp};
	my $rowno = $arg{rowno};
	my @vals;
	if ( $rp )
	{
		for my $field (@$fields)
		{
			confess "row_value_join: unknown field $field" unless defined $rp->{$field};
			push @vals, $rp->{$field};
		}
	}
	elsif ( defined $rowno )
	{
		my $ds = $arg{datasource} || $this->{datasource} || confess;
		my $row = $ds->{row}->[$rowno] || return '';
		for my $field (@$fields)
		{
			my $colno = $ds->{name2colno}->{$field};
			confess "row_value_join: unknown column $field" unless defined $colno;
			push @vals, $row->[$colno] if $row->[$colno];
		}
	}
	else
	{
		confess "expected rp or rowno";
	}
	my $s = join($sep, @vals);
	if ( $sep eq ' ' || $sep eq ', ' )
	{
		$s =~ s/\s+/ /g;
		$s =~ s/^\s+//;
		$s =~ s/\s+$//;
	}
	$s;
}

sub value_join
{
	my $this = shift;
	my $sep = shift || ' ';
	my @vals;
	for (@_)
	{
		push @vals, $_ if defined $_;
	}
	my $s = join($sep, @vals);
	if ( $sep eq ' ' )
	{
		$s =~ s/\s+/ /g;
		$s =~ s/^\s+//;
		$s =~ s/\s+$//;
	}
	$s;
}

#- helper functions

sub search_substr
{
	my $this = shift;
	my %arg = ( iv => undef, field => undef, value => undef, wherestrptr => undef, whereptr => undef, dbptr => undef, dbdef => undef, fromptr => undef, @_ );
	push @{$arg{whereptr}},
	    $arg{field} => { 'like', '%'.$arg{value}.'%' }
	  if $arg{value};
	1;
}


1;
__END__

=head1 NAME

App::ListBase

=head1 DESCRIPTION

List handler base class

=head1 AUTHOR

Peter Edwards of Dragonstaff Limited <peter@dragonstaff.co.uk>

=head1 COPYRIGHT

Copyright (c) 2007 Dragonstaff Limited.

All rights reserved.

=cut
