package App::Web;

use strict;
use warnings;

use Carp;
use CGI;
use CGI::Session;
use Data::Dumper;
use Config::General;
use Template;
use Mail::SendVarious;
use Image::Thumbnail;
use Time::HiRes qw/ gettimeofday tv_interval /;

#use Apache2::ServerUtil();

use App::Controller;
use App::Model;

use vars qw(%c $session $cgi);
*c = \*app::c;
*session = \$app::session;
*cgi = \$app::cgi;


sub run
{
  my $class = shift;
  my $this = {};
  bless $this, $class;

  if (exists $main::ENV{MOD_PERL})
  {
    #Apache2::ServerUtil->server->pool->cleanup_register(\&end);
  }
  $this->init;
 
  #print Dumper(%c);
  #print Dumper($session);
  #print Dumper($cgi);

  $this->add_bench('controller');
  $this->controller;

  $this->add_bench('model');
  $this->model;

  $this->add_bench('view');
  $this->view;

  $this->add_bench('send_response');
  $this->send_response;

  $this->add_bench('end');
  $this->end;

  $this;
}

sub init
{
  my $this = shift;
  
  # global variables and configuration
  # TODO split between persistent/cleared each run
  %app::c = () unless defined %app::c;
  $app::cgi = undef;
  $app::session = undef;
  *c = \*app::c;
  *session = \$app::session;
  *cgi = \$app::cgi;

  $c{bench} = {
    start   => [gettimeofday()],
    points  => [],
    end     => [gettimeofday()],
    };

  #
  $this->load_config;

  # running under mod_perl or CGI server mode
  $c{mod_perl} = exists $main::ENV{MOD_PERL} ? 1 : 0;
  
  # allow for interactive debugging
  $c{interactive_user} = ( -t STDERR && $main::ENV{SHELL} ) ? 1 : 0;

  # CGI form processor
  my $cgidebug = $c{interactive_user} && !$main::ENV{NOCGIDEBUG} ? 1 : 0;
  # URI path will map to controller
  if ( $cgidebug )
  {
    print STDERR "(offline mode: enter PATH_INFO)\n" if -t STDERR;
    chomp($c{path_info} = (<>||""));
  }
  else
  {
    $c{path_info} = (substr(($main::ENV{PATH_INFO}||' '),1,255)||'');
  }
  $c{hosturl} = 'http://'.($main::ENV{HTTP_HOST}||'localhost');
  #print STDERR Dumper(%main::ENV);

  # form params will be processed by controller
  $CGI::DEBUG = $cgidebug ? 2 : 0;
  $app::cgi = new CGI;

  $this->debuglog( 'path_info: ', $c{path_info} );

  # model
  unless ( $c{schema} )
  {
    $c{schema} = "App::Model";
    $c{schema}->load( $c{db_dsn}, $c{db_user}, $c{db_pass} );
  }

  # session
  my $sess_dsn = $c{cgi_session_dsn}||"driver:mysql";
  my $sess_id  = $c{cgi_session_id}||"id:md5";
  my $sess_dsn_args = {
    DataSource  => ($c{db_dsn}||confess),
    User        => ($c{db_user}||confess),
    Password    => ($c{db_pass}||confess)
  };
  my $sid = undef;
  if ( $c{interactive_user} )
  {
    $sid ||= $ENV{SESSDEBUGSID} || $ENV{USER}|| getpwuid($<);
    $app::session = new CGI::Session($sess_dsn.';'.'id:static', $sid, $sess_dsn_args ) || confess "init: ".CGI::Session->errstr;
    $app::session->{_QUERY} = $app::cgi; # no easy way to get an id:static into CGI::Session along with an external CGI object
  }
  else
  {
    $app::session = new CGI::Session($sess_dsn.';'.$sess_id, $app::cgi, $sess_dsn_args ) || confess "init: ".CGI::Session->errstr;
  }
  $this->add_bench('init_session');
  # set up session data areas
  $this->init_session;

  # set up response area
  $this->init_response;

  $this->init_view;
  
  1;
}

sub load_config
{
  my $this = shift;
  my $bin = FindBin::Real::Bin();
  my $path = $bin."/app.conf"; # chdir() done in app.pl
  my $conf = Config::General->new($path)||confess "can't read config file $path: $!";
  my %conf = $conf->getall;
  %c = ( %c, %conf );
  $c{BINROOT} = $bin;
  $c{LISTDEFROOT} = $bin . "/listdef";
  $c{FORMDEFROOT} = $bin . "/formdef";
  $c{TEMPLATE_INCLUDE_PATH} = $bin . '/template'; # :alternatepath1:alternatepath2
}

sub add_bench
{
  my $this = shift;
  my $bench_name = shift||confess;
  push @{$c{bench}->{points}}, { name => $bench_name, tv => [gettimeofday()] };
}

sub end
{
  my $this = shift;

  $session && $session->flush;

  # benchmark
  if ( $this && $c{benchmark_enabled} )
  {    
    $c{bench}->{end} = [gettimeofday()];
    print STDERR 'Bench run time: ', tv_interval($c{bench}->{start}, $c{bench}->{end}), "\r\n";
    $this->debuglog( 'Bench run time: ', tv_interval( $c{bench}->{start}, $c{bench}->{end} ) );
    for (@{$c{bench}->{points}})
    {
      print STDERR 'Bench '.$_->{name}.': ', tv_interval( $c{bench}->{start}, $_->{tv} ), "\r\n";
      #$this->debuglog( 'Bench '.$_->{name}.': ', tv_interval( $c{bench}->{start}, $_->{tv} ) );
    }  
  }

  $app::session = undef;
  $app::cgi = undef;
  # persist only what is necessary
  %app::c = (
    schema    => $app::c{schema}, # DBIx::Class database schema
  );
  #$app::db = undef;
  #return Apache2::Const::OK;
}

sub DESTROY { end(); }

sub debuglog
{
  my $this = shift;
  return unless $c{debug_log};
  unless ( open(OUTPUT, '>>'.$c{debug_log}) )
  {
    warn "cannot open debug_log: $c{debug_log}: $!";
    return 0;
  }
  unless ( $c{debug_log_started} )
  {
    $c{debug_log_started} = 1;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    print OUTPUT sprintf("+++ %04d-%02d-%02d %02d:%02d:%02d +++\r\n", $year+1900,$mon+1,$mday,$hour,$min,$sec);
  }  
  for (@_)
  {
    print OUTPUT $_;
  }
  print OUTPUT "\r\n";
  close OUTPUT;
  
}

sub user_fields
{
  qw/ id login location name email_optout is_admin /;
}

sub set_session_user
{
  my $this = shift;
  $this->set_session_user_from_rs(undef, 1);
}

sub set_session_user_from_rs
{
  my $this = shift;
  my $rsuser = shift;
  my $default_from_session = shift;
  my $user = $session->param('user');
  if ( !$rsuser && $default_from_session && $user->{is_logged_in} && $user->{login} )
  {
    $rsuser = $c{schema}->resultset('User')->find({ login => $user->{login} });
  }
  for ( user_fields() )
  {
    $user->{$_} = $rsuser ? $rsuser->get_column($_) : '';
  }
  if ( $rsuser )
  {
    if ( $default_from_session )
    {
      $user->{is_logged_in} ||= 0;
    }
    else
    {
      $user->{is_logged_in} = 1;
    }
  }
  else
  {  
    $user->{is_logged_in} = 0;
  }
  $session->param('user', $user);
}

sub init_session
{
  my $this = shift;

  # get and set current user if any from session stash
  $this->set_session_user;

  # save cgi params in session for debugging
  my $cgidebug = $session->param('cgi');
  $cgidebug->{query_string} = $cgi->query_string;
  $cgidebug->{path_info}    = $c{path_info};
  $cgidebug->{remote_addr}  = $session->remote_addr();
  $cgidebug->{debugparams}  = $c{path_info} . "\n" . ($cgi->param('POSTDATA')||'');
  #$user->{cgi}->{dump} = Dumper($cgi);
  $session->param('cgi', $cgidebug);

  # log each transaction run
  my $user = $session->param('user');
  $c{schema}->resultset('AccessLog')->create({
    session_id    => $session->id,
    user_id       => $user->{id},
    login         => $user->{login},
    remote_addr   => $session->remote_addr(),
    path_info     => $c{path_info},
    query_string  => $cgi->query_string,
    postdata      => ($cgi->param('POSTDATA')||undef),
    });

  1;
}

#=== View

sub init_view
{
  my $this = shift;
  $c{baseurl} = $c{mod_perl} ? '/lpf/app/' : '/lpf/app/';
  $c{view}->{show_menu} = 1;
#print STDERR "running in $c{mod_perl} mode\r\n";

  # retrieve messages if any from session following a redirect ready to display, then clear them from session
  my $message = $session->param('message')||[];
  $session->param('message', []);
  push @{$c{v}->{message}}, @$message;
}

#=== Responses - either calculated or from named template optionally inside wrapper templates

sub init_response
{
  my $this = shift;
  $c{response} = {
      code => 200,
      redirect => '',
      content => '',
      header => {},
      title => 'Sample App',
      };
  $c{template} = '';
  $c{wraptemplates} = [ 'standard.tt2', 'xhtml.tt2' ];
  $c{v} = {}; # template vars
  $c{response}->{header}->{'Set-Cookie'} = "".$session->cookie;
  $c{response}->{header}->{'Connection'} = "close";
  $this->set_content_type('text/html; charset=iso-8859-1');
}

=head1

HTTP/1.1 200
Date: Wed, 22 Nov 2006 10:26:46 GMT
Server: Apache/2.0.54 (Fedora)
Content-Length: 4725
Connection: close
Content-Type: text/html; charset=iso-8859-1

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">

<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">

<head>

=cut

sub header_xhtml
{
  return <<EOM ;
\015\012<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
EOM
}

sub set_default_response
{
  my $this = shift;
  $c{response}->{code} = 404;
  $c{response}->{content} = header_xhtml() . '<head><title>Default Reponse</title></head><body>This is the default reponse from the web server. Someone has probably made a programming mistake.</body></html>';
  $this->set_content_type('text/html; charset=iso-8859-1');
  $c{response}->{title} = '\015\012Web';
  #push @{$c{response}->{mimeheader}}, ();
}

sub set_error_response
{
  my $this = shift;
  my $errmsg = shift || confess "expected errmsg";
  my $code = shift || 404;
  $c{response}->{code} = $code;
  $c{response}->{content} = header_xhtml() . '<head><title>Error</title></head><body>An error has occurred in the Sample App web server<br>'.$errmsg.'</body></html>';
  $this->unset_template;
}

sub set_xhtml_response
{
  my $this = shift;
  my $resp = shift||confess;
  $this->set_content_type('text/html; charset=iso-8859-1');
  $c{response}->{code} = 200;
  $c{response}->{content} = "\015\012$resp\015\012";
  $this->unset_template;
}

sub set_wrapped_xhtml_response
{
  my $this = shift;
  my $resp = shift||confess;
  $this->set_xhtml_response(
    header_xhtml()
  . "<head><title>".($c{response}->{title}||'Web')."</title></head><body>\015\012"
  . $resp
  . "\015\012</body></html>\015\012"
  );
}

sub set_xml_response
{
  my $this = shift;
  my $resp = shift||confess;
  $this->set_content_type('text/xml; charset=iso-8859-1');
  $c{response}->{code} = 200;
  $c{response}->{content} = "\015\012$resp\n";
  $this->unset_template;
  1;
  
  # print $cgi->header(-type => 'text/xml'),
  #      $cgi->start_html(-declare_xml, -dtd);
  #
  # gives
  #
  # Content-Type: text/xml; charset=ISO-8859-1
  #
  #<?xml version="1.0" encoding="iso-8859-1"?>
  #<!DOCTYPE html
  #        PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  #         "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
  #<html xmlns="http://www.w3.org/1999/xhtml" lang="en-US" xml:lang="en-US">
  #<head>
  #<title>Untitled Document</title>
  #</head>
  #<body>
}

sub set_csv_response
{
  my $this = shift;
  my $resp = shift;
  confess unless defined $resp;
  my $filename = shift||'download';
  $this->set_content_type('text/csv; charset=iso-8859-1');
  $c{response}->{header}->{'Content-disposition'} = 'attachment; filename='.$filename.'.csv';
  $c{response}->{code} = 200;
  $c{response}->{content} = "\015\012$resp\n";
  $this->unset_template;
  1;
}

sub set_content_type
{
  my $this = shift;
  my $type = shift||confess;
  $c{response}->{header}->{'Content-Type'} = $type;
}

sub unset_template
{
  my $this = shift;
  $c{template} = '';
  $c{wraptemplates} = [];
}

sub redirect
{
  my $this = shift;
  my $url = shift||confess;
  # Firefox expects full location URL not relative
  unless ( $url =~ m/^http:/ )
  {
    $url =~ s|^\/||;
    $url = $c{hosturl}.$c{baseurl}.$url;
  }  
  $session->param('message', ($c{v}->{message}||[]));

  # 302 redirect following a non-cacheable add or update (should be 303 according to W3C? 302 is historic but seems to be standard)
  print "Status: 302 Moved\015\012";
  print "Location: ".$url."\015\012";
  print "Set-Cookie: ".$session->cookie."\015\012" if $session->cookie;
  #print "Connection: close\015\012";
  #print "Content-Type: text/html; charset=iso-8859-1\n";
  print "\015\012";
  exit(0);
}

#

sub send_response
{
  my $this = shift;

  $this->debuglog( 'send_response: ' );
   
  return if $c{done_response};

  # code
  #print "HTTP/1.1 ".$c{response}->{code}."\015\012";

  my $resp = $c{response};

  # MIME header
  while ( my ($k, $v) = each %{$resp->{header}} )
  {
    $this->debuglog( "header: $k: $v" );
    print "$k: $v\015\012";
  }

  # content
  $this->debuglog( 'content: '.substr($resp->{content},0,30).' ... '.substr($resp->{content},length($resp->{content})-30,30)."\r\n" );
  for (my $i = 0; $i < length($c{response}->{content}); $i += 1024)
  {
    # need to chunk the output or mod_perl barfs
    print unpack("x$i a1024", $resp->{content});
  }
}

#=== Model

sub model
{
  my $this = shift;
}
#=== View

# view handling
sub view
{
  my $this = shift;
  return unless $c{template} || @{$c{wraptemplates}};
  my $t = Template->new({
    INCLUDE_PATH    => $c{TEMPLATE_INCLUDE_PATH}, # TODO from config
		PLUGINS         => { W => "App::Widget" },
  });

  # basic
  $c{v}->{baseurl} = $c{baseurl};
  $c{v}->{session_id} = $session->id;
  $c{v}->{formstart} = '<form name="Form" method="post" action="'.($c{v}->{formaction}||$c{path_info}||$c{baseurl}).'" enctype="text.plain" '.($c{v}->{formextra}||'').'>'
    unless defined $c{v}->{formstart};
  $c{v}->{formend} = '<input type="hidden" name="button_name" value="" /><input type="hidden" name="button_value" value="" /></form>' unless defined $c{v}->{formend};
  # debug
  $c{v}->{env} = \%main::ENV;
  # user
  #$c{v}->{user} = $session->param('user');
  my $sessuser = $session->param('user');
  $c{v}->{sess_user} = $sessuser;
  # session
  
  # permissions
  my $is_logged_in = $sessuser->{is_logged_in} ? 1 : 0;
  $c{v}->{perm} = {
    admin   => ($is_logged_in && ($sessuser->{is_admin} eq 'yes') ? 1 : 0),
  };

  # wrapper view settings
  $c{v}->{show_menu} = $c{view}->{show_menu};

  my $v = '';
  if ( $c{template} )
  {
    $t->process($c{template}, $c{v}, \$v)
      || return $this->set_error_response("template error: ".$t->error());
  }
  for my $tmplname (@{$c{wraptemplates}})
  {
    $c{v}->{wraptemplatecontent} = $v;
    $v = '';
    $t->process($tmplname, $c{v}, \$v)
      || return $this->set_error_response("wraptemplate error: ".$t->error());
  }
  $c{response}->{content} = $v;
}

#=== Controller

# main controller entry point
sub controller
{
  my $this = shift;
  App::Controller->run( web => $this );
}  

#=== utilities

sub read_structure_from_file
{
  my $this = shift;
  my $path = shift||confess;
  local (*INPUT, $/);
  open (INPUT, $path) || return {};
  my $input = <INPUT>;
  my $data = '';
  eval ( $input );
  return $data;
}

sub write_structure_from_file
{
  my $this = shift;
  my $path = shift||confess;
  my $structure = shift;
  my $data = Data::Dumper->new ( [$structure], [ 'data' ] );
  my $output = $data->Dump;
  return 0 unless open(OUTPUT, '>'.$path);
  print OUTPUT $output;
  close OUTPUT;
  return 1;
}

sub read_file
{
  my $this = shift;
  my $path = shift||confess;
  local (*INPUT, $/);
  open (INPUT, $path) || return '';
  <INPUT>;
}

sub write_file
{
  my $this = shift;
  my $path = shift||confess;
  my $s = shift;
  return 0 unless open(OUTPUT, '>'.$path);
  print OUTPUT $s;
  close OUTPUT; # || confess; # kinda doomed anyway if we run out of space to write...
  return 1;
}

sub url_encode
{
  my $s = shift || return '';
  $s =~ s/([^A-Za-z0-9])/sprintf("%%%02X", ord($1))/seg;
  $s;
}

sub url_decode
{
  my $s = shift || return '';
  $s =~ s/\%([A-Fa-f0-9]{2})/pack('C', hex($1))/seg;
  $s;
}

sub escape_html
{
  #my $this = shift;
  my $s = shift;
  return '' unless defined $s;
  my %map = (
    "\<"      => "&lt;",
    "\>"      => "&gt;",
    "\&"      => "&amp;",
    "\""      => "&quot;",
    "\@"      => "&#064;",
    "\'"      => "&#146;",
    );
  no warnings;
  $s =~ s{ ( [\Q<>&"'@\E] ) } { $map{$1} if $1 }gsex;
  use warnings;
  $s;
}

sub send_mail
{
  my $this = shift;
  my %arg = @_;
  my $from = $arg{from}||confess "send_email: no from address";
  my $from_name = $arg{from_name}||'';
  my $to = $arg{to}||confess "send_email: no to address";
  my $cc = $arg{cc}||'';
  my $to_name = $arg{to_name}||'';
  my $subject = $arg{subject}||confess "send_email: no subject";
  my $body = $arg{body}||'';
  my $header = $arg{header}||'';
  my $xheader = $arg{xheader}||'';
  my $message = $arg{message}||'';
  my $message_content = $arg{message_content};
  my $host = $arg{host}||undef;

  if ( $message_content )
  {
    $xheader .= "Content-Type: $message_content; charset=\"iso-8859-1\"\n\n";
  }
  else
  {
    $xheader .= "Content-Type: text/html; charset=\"iso-8859-1\"\n\n";
  }

  # log sent email
  if ( $c{mail_log} )
  {
    $c{schema}->resultset('MailLog')->create({
      mail_to   => $to,
      mail_from => $from,
      subject   => $subject,
      body      => $body,
      });
  }

  # send email
  my $rc = sendmail(
    from => $from,
    From => $from_name,
    to => $to,
    cc => $cc,
    subject => $subject,
    header => $header,
    xheader => $xheader,
    body => $body,
    message => $message,
    debuglogger => sub {},
    );
  my $msg = $rc ? 'Good' : $Mail::SendVarious::mail_error;

  ($rc, $msg);
}

sub DEFAULT_THUMBNAIL_SIZE { '76x56' };

sub image_to_thumbnail
{
  my $this = shift;
  my %arg = @_;
  my $filein  = $arg{filein} || confess __PACKAGE__.": expected filein arg";
  my $fileout = $arg{fileout} || confess __PACKAGE__.": expected fileout arg";
  my $size    = $arg{size} || DEFAULT_THUMBNAIL_SIZE;

  unlink($fileout);
  ! -r $fileout || confess __PACKAGE__.": couldn't remove fileout $fileout: $!";

  my $t = Image::Thumbnail->new(
    size        => $size,
    inputpath   => $filein,
    outputpath  => $fileout,
    create      => 1,
    );
  return ($t->{error} ? (0,"failed to convert image to thumbnail: $filein: $t->{error}") : (1,''));
}


__END__

=head1 NAME

App::Web

=head1 DESCRIPTION

Sample Lightweight Perl Framework Application

=head1 AUTHOR

Peter Edwards of Dragonstaff Limited <peter@dragonstaff.co.uk>

=head1 COPYRIGHT

Copyright (c) 2007 Dragonstaff Limited.

All rights reserved.

=cut
